'Go run the test suite. Remove this line and the TestSuite to include BCDLiB24Digits
'into your own code. 
 GOTO TestSuite   
 
 
'**************************************************************************
'BCD Library for MMBasic. BCDLib24DigitsV3.BAS
'Ver 3, 19 August 2014 
'
'This Version has been hand OPTIMISED for 24 decimal digits. Do not use with
'other digit lengths.
'
'The algorithms used in this Library typically operate on sextets of digits at a time.
'
'
'(c) Peter Carnegie 2014 with thanks to TassyJim from the BackShed
'who inspired this library with his BigInt Library, and to JohnS also from
'the BackShed who had the insight that one could safely use sextets of digits
'per numeric var thereby hugely improving performance.
'
'This library is NOT Re-entrant, eg should not be used within ISR's
'
'Library uses 3 arrays, BCD.A(),BCD.B(),BCD.C()
'BCD.A(),BCD.B(),BCD.C() will be declared as GLOBAL vars by BCD.InitLib()
'These arrays are treated as BCD registers capable
'of holding up to 24 decimal digit numbers.
'
'Numbers are stored in unpacked form, 6 decimal digits per
'array entry, in little endian form.
'eg 123456789 is stored in BCD.A() as
'BCD.A(0)=456789, BCD.A(1)=000123, BCD.A(2)=000000
'
'Algorithms processes Sextets of digits at a time reducing
'processing times, eg on a uMite at 48MHz, an 18 digit BCD.AddAB2C
'operation takes about 3 mS.
'
'Generally, register BCD.C() will contain the results of
'two part operations, eg BCD.AddAB2C takes BCD.A() + BCD.B()
'and returns the result in BCD.C().
'
'
'
'The list of functions/subs in the library include
'
'  BCD.InitLib()           	MUST be called first. Creates Global BCD Lib Vars, BCD.A/B/C(), BCD.Width,BCD.NumDigits,BCD.Size
'                         	BCD.Width will be set to 6, BCD.Size will be set to 4, BCD.NumDigits will be set to 24
'
'  BCD.AddAB2C				BCD.C()=BCD.A()+BCD.B()
'  BCD.AddAB2A				BCD.A()=BCD.A()+BCD.B()
'  BCD.SubAB2C				BCD.C()=BCD.A()-BCD.B()
'  BCD.SubAB2A				BCD.A()=BCD.A()-BCD.B()
'  BCD.IncA(Inc)			BCD.A()=BCD.A()+Inc  Note Inc is a SINGLE
'  BCD.DecA(Dec)			BCD.A()=BCD.A()-Dec  Note Dec is a SINGLE
'  BCD.MulAB2C				BCD.C()=BCD.A()*BCD.B()
'  BCD.DivAB2C				BCD.C() = BCD.A()\BCD.B()
'
'Utility Routines******************
'
'  BCD.Clr(Dest,Width)		Dest(0..Width-1)=0, Width defaults to 4 ie 24 dedimal digits
'  BCD.Load(Dest,String)	Dest() is loaded with contents of string, 6 digits per array element
'
'
'  BCD.LPad$(A$,ML)			Pad a string with leading zeroes to make a string ML chars long
'
'  BCD.A2Bin				    BCD.C() = BCD to Binary conversion of BCD.A() in little endian format
'  BCD.A2Asc$()				  Returns String representation of BCD.A()
'
'  BCD.Print(Src,Mode,NumDigits)	Print a BCD register to console,Src is first element of BCDregister, eg FREQ(0)
'
'  BCD.Copy(Src,Dest,NumDigits)		Copy a BCD number from source to dest, Src and Dest are first element of BCD register, eg BCD.A(0)
'
'**************************************************************************


  
'=========================================================
'
'Initialise the BCD Library
'MUST be called BEFORE calling any other BCD Lib routines
'Creates Global BCD Lib Vars, BCD.A/B/C() 
'    BCD.Size Contains number of Numeric Vars per BCD Register, always 4
'    BCD.Width contains the number of decimal digits per Numeric Var, Always 6
'    BCD.NumDigits contains the number of digits per BCD Register, al;ways 24
'
SUB BCD.InitLib()
  
  DIM BCD.Width,BCD.Size,BCD.NumDigits
  
  'Set precision to 24 decimal digits
  BCD.NumDigits=24
    
  'Number of Digits per Numeric Var
  BCD.Width=6
    
  'Number of Numeric vars per BCD register
  'Each Numeric var holds BCD.Width digits
  BCD.Size=(BCD.NumDigits\BCD.Width)
    
  'NB BCD.C is TWICE as long as the others to allow for Multiplication
  DIM BCD.A(BCD.Size-1),BCD.B(BCD.Size-1),BCD.C(BCD.Size*2)
 
END SUB
  
'=========================================================
'BCD.C() = BCD.A() + BCD.B()
'
'BCD.A(0..3),BCD.B(0..3),BCD.C(0..7) Single Arrays
'Each Element stores 6 digits of the overall number
'The numbers are stored in little Endian format
'
'Algorithm works on sextet of digits at a time
'Optimised for 24 digits
'
SUB BCD.AddAB2C()
  LOCAL Cy,I
  
  'Clear the BCD C Register
  BCD.C(0)=0:BCD.C(1)=0:BCD.C(2)=0:BCD.C(3)=0
  BCD.C(4)=0:BCD.C(5)=0:BCD.C(6)=0:BCD.C(7)=0:
  
  I=BCD.A(0)+BCD.B(0)
  Cy=I\&HF4240
  BCD.C(0)=I MOD &HF4240
  
  I=BCD.A(1)+BCD.B(1)+Cy
  Cy=I\&HF4240
  BCD.C(1)=I MOD &HF4240
  
  I=BCD.A(2)+BCD.B(2)+Cy
  Cy=I\&HF4240
  BCD.C(2)=I MOD &HF4240
  
  I=BCD.A(3)+BCD.B(3)+Cy
  BCD.C(3)=I MOD &HF4240
  
END SUB

'=========================================================
'BCD.A() = BCD.A() + BCD.B()
'
'BCD.A(0..3),BCD.B(0..3) Single Arrays
'Each Element stores 6 digits of the overall number
'The numbers are stored in little Endian format
'
'Algorithm works on sextet of digits at a time
'Optimised for 24 digits
'
SUB BCD.AddAB2A()
  LOCAL Cy,I
  
  I=BCD.A(0)+BCD.B(0)
  Cy=I\&HF4240
  BCD.A(0)=I MOD &HF4240
  
  I=BCD.A(1)+BCD.B(1)+Cy
  Cy=I\&HF4240
  BCD.A(1)=I MOD &HF4240
  
  I=BCD.A(2)+BCD.B(2)+Cy
  Cy=I\&HF4240
  BCD.A(2)=I MOD &HF4240
  
  I=BCD.A(3)+BCD.B(3)+Cy
  BCD.A(3)=I MOD &HF4240
  
END SUB
  
 
'=========================================================
'
'Subtract BCD.B from BCD.A
'Result in BCD.C
'
'Optimised for 24 digits
'
'
SUB BCD.SubAB2C()
  
  LOCAL I,J,Cy
  
  BCD.C(0)=BCD.A(0):BCD.C(1)=BCD.A(1):BCD.C(2)=BCD.A(2):BCD.C(3)=BCD.A(3):
  
  BCD.C(4)=0:BCD.C(5)=0:BCD.C(6)=0:BCD.C(7)=0
  
  J=BCD.C(0)+&HF4240-BCD.B(0):Cy=1-(J\&HF4240):BCD.C(0)=J MOD &HF4240
  
  J=BCD.C(1)+&HF4240-BCD.B(1)-CY:Cy=1-(J\&HF4240):BCD.C(1)=J MOD &HF4240
  
  J=BCD.C(2)+&HF4240-BCD.B(2)-CY:Cy=1-(J\&HF4240):BCD.C(2)=J MOD &HF4240
  
  J=BCD.C(3)+&HF4240-BCD.B(3)-CY:BCD.C(3)=J MOD &HF4240
  
END SUB
  
'=========================================================
'
'Subtract BCD.B from BCD.A
'Result in BCD.A
'
'Optimised for 24 digits
'
SUB BCD.SubAB2A()
  
  LOCAL Cy,I
  
  I=BCD.A(0)+&HF4240-BCD.B(0):Cy=1-(I\&HF4240):BCD.A(0)=I MOD &HF4240
  
  I=BCD.A(1)+&HF4240-BCD.B(1)-CY:Cy=1-(I\&HF4240):BCD.A(1)=I MOD &HF4240
  
  I=BCD.A(2)+&HF4240-BCD.B(2)-CY:Cy=1-(I\&HF4240):BCD.A(2)=I MOD &HF4240
  
  I=BCD.A(3)+&HF4240-BCD.B(3)-CY:BCD.A(3)=I MOD &HF4240
  
END SUB
  
'=========================================================
'
'Increment BCD.A() by Inc
'
'NB Inc is a MMBasic single
'
'Optimised for 24 decimal digits
'
SUB BCD.IncA(Inc)
  
  LOCAL Cy

  BCD.A(0)=BCD.A(0)+Inc
  IF BCD.A(0)<&HF4240 THEN
    EXIT SUB
  ELSE
    Cy=BCD.A(0)\&HF4240
    BCD.A(0)=BCD.A(0) MOD &HF4240
  ENDIF
  
  BCD.A(1)=BCD.A(1)+Cy
  IF BCD.A(1)<&HF4240 THEN
    EXIT SUB
  ELSE
    Cy=BCD.A(1)\&HF4240
    BCD.A(1)=BCD.A(1) MOD &HF4240
  ENDIF
  
  BCD.A(2)=BCD.A(2)+Cy
  IF BCD.A(2)<&HF4240 THEN
    EXIT SUB
  ELSE
    Cy=BCD.A(2)\&HF4240
    BCD.A(2)=BCD.A(2) MOD &HF4240
  ENDIF
  
  BCD.A(3)=BCD.A(3)+Cy
  IF BCD.A(3)<&HF4240 THEN
    EXIT SUB
  ELSE
    BCD.A(3)=BCD.A(3) MOD &HF4240
  ENDIF
    
END SUB
  
  
'=========================================================
'
'Decrement BCD.A() by Dec
'
'NB Dec is an MMBasic Numeric Var
'
' This SUB is optimised for 24 digits only
'
SUB BCD.DecA(Dec)
  
  LOCAL I, Bw
  
  Bw=Dec
  
  I=BCD.A(0)+&HF4240-Dec
  BW=1-(I\&HF4240)
  BCD.A(0)=I MOD &HF4240
  IF Bw=0 THEN EXIT SUB
  
  
  I=BCD.A(1)+&HF4240-BW
  BW=1-(I\&HF4240)
  BCD.A(1)=I MOD &HF4240
  IF Bw=0 THEN EXIT SUB
  
  I=BCD.A(2)+&HF4240-Bw
  BW=1-(I\&HF4240)
  BCD.A(2)=I MOD &HF4240
  IF Bw=0 THEN EXIT SUB
  
  I=BCD.A(3)+&HF4240-BW
  BCD.A(3)=I MOD &HF4240

END SUB
  
'=========================================================
'BCD.C()=BCD.A() x BCD.B()
'This algorithm works on Sextets of digits
'For optimum speed, make sure the shortest multiplier is in BCD.B()
'
'BCD.MulAB2C reorganises data and then process in triplets, and then repacks the result
'
SUB BCD.MulAB2C
  
  LOCAL Cy,I,J,K,L,M,N,Ele
  
  'These are "shadow" BCD registers used for MUL & Div which are carried out in triplets
  LOCAL BCD._A((BCD.Size*2)-1),BCD._B((BCD.Size*2)-1),BCD._C(BCD.Size*4)
  
  'Unpack sextets into BCD._A() and BCD._B
  J=0
  FOR I=0 TO BCD.Size-1
    BCD._A(J)=BCD.A(I) MOD 1000
    BCD._A(J+1)=BCD.A(I) \ 1000
    
    BCD._B(J)=BCD.B(I) MOD 1000
    BCD._B(J+1)=BCD.B(I) \ 1000
    
    J=J+2
  NEXT I
  
  'Array for the intermediate multiplication
  LOCAL C1(BCD.Size*4)
  
  Ele=(BCD.Size*2)-1
  
  'Set default length
  M=Ele
  'Find length of BCD.A()
  FOR I=Ele TO 0 STEP -1
    IF BCD._A(I)<>0 THEN
      M=I
      EXIT FOR
    ENDIF
  NEXT I
  
  'Set default length
  N=Ele
  'Find Length of BCD.B()
  FOR I=Ele TO 0 STEP -1
    IF BCD._B(I)<>0 THEN
      N=I
      EXIT FOR
    ENDIF
  NEXT I
  
  'Now Do the Long Multiplication of the two numbers
  L=0
  K=0
  
  FOR I=0 TO M+1
    Cy=0
    FOR J=0 TO N+1
      
      C1(J)=(BCD._A(I)*BCD._B(J))+Cy
      
      IF C1(J)>&H3E7 THEN       '&H3E7=999
        Cy=C1(J)\&H3E8          '&H3E8=1000
        C1(J)=C1(J) MOD &H3E8
      ELSE
        Cy=0
      ENDIF
      
    NEXT J
    
    'Add the partial result to the overall total
    'Add each element of C() to D() result into D()
    Cy = 0
    
    FOR J=0 TO Ele
      K=J+L
      BCD._C(K)=BCD._C(K)+C1(J)+Cy
      IF BCD._C(K)>&H3E7 THEN
        Cy=BCD._C(K)\&H3E8 
        BCD._C(K)=BCD._C(K) MOD &H3E8 
      ELSE
        Cy =0
      ENDIF
    NEXT J
    
    L=L+1
    
  NEXT I
  
  'Repack the Result from BCD._C() into BCD.C()
  J=0
  FOR I=0 TO (BCD.Size*2)-1 STEP 2
    BCD.C(J)=BCD._C(I) + BCD._C(I+1)*&H3E8 
    J=J+1
  NEXT I
  
END SUB
  
 
'=========================================================
' Dividend in BCD.A()
' Divisor  in DCDB()
' Result   in BCD.C()
'
' BCD.C() = BCD.A() \ BCD.B()
'
' Usage :
'     BCD.Load BCD.A(0),"123456780"
'     BCD.Load BCD.B(0),"123"
'     BCD.Clr  BCD.C(C),1
'     BCD.DivAB2C
'     BCD.Print BCD.C(0),0,1
'
' BCD.DivAB2C is adapted from TassyJim's BigInt Div Subroutine
' This sub unpacks and re-packs sextet sized BCD registers to and from triplet
' sized registers. The actual work is carried out in Triplets
'
' A DIVIDE operation takes about 400 mS
'
'
SUB BCD.DivAB2C()
  
  LOCAL D,I,J,K,M,N
  LOCAL Bw,Cy             'Borrow and Carry
  
  LOCAL A3(BCD.Size*2),A3L 'Dividend
  LOCAL A4(BCD.Size*2)     'Divisor 10s Left Shifted
  LOCAL A5(BCD.Size*2)     'Remainder
  LOCAL B3(BCD.Size*2)     'Divisor
  LOCAL C1(BCD.Size*2)     'Try
  LOCAL C2(BCD.Size*4),C2L  'Captures single digit intermediate results
  LOCAL C3(BCD.Size*2)     'Used to compare Dividend - Try
  LOCAL Diff(BCD.Size*2)   'Difference of Dividend - Try
  
  'These are "shadow" BCD registers used for MUL & Div which are carried out in triplets
  LOCAL BCD._A((BCD.Size*2)-1),BCD._B((BCD.Size*2)-1),BCD._C(BCD.Size*4)
  
  'Unpack Sextets into BCD._A() and BCD._B
  J=0
  FOR I=0 TO BCD.Size-1
    BCD._A(J)=BCD.A(I) MOD &H3E8        '&H3E8 =1000
    BCD._A(J+1)=BCD.A(I)\&H3E8 
    
    BCD._B(J)=BCD.B(I) MOD &H3E8 
    BCD._B(J+1)=BCD.B(I)\&H3E8 
    
    J=J+2
  NEXT I
  
'  FOR I=0 TO (BCD.Size*4)-1
'    BCD._C(I)=0
'  NEXT I
  
  Ele=(BCD.Size*2)-1
  
  'Find First Non-Zero triplet in BCD.A()
  M=7
  FOR I=Ele TO 0 STEP -1
    IF BCD._A(I)<>0 THEN
      M=I
      EXIT FOR
    ENDIF
  NEXT I
  
  'Save number of Triplets of BCD.A() to process
  A3L=M
  'Copy BCD.A() into working storage
  FOR I=M TO 0 STEP -1
    A3(I)=BCD._A(I)
  NEXT I
  
  'A3(M) contains first non-zero triplet, but which digit
  'within A3(M) is non zero
  J=0
  IF A3(M)>99 THEN
    J=3
  ELSE
    IF A3(M)>9 THEN J=2
    IF A3(M)<10 THEN J=1
  ENDIF
  'Number of significant digits in BCD.A()
  M=(M*3)+J
  
  'Find out how many triplets in the Divisor
  N=ele
  FOR I=ele TO 0 STEP -1
    IF BCD._B(I)<>0 THEN
      N=I
      EXIT FOR
    ENDIF
  NEXT I
  
  'N now contains Number of Triplets in the Divisor
  'Copy Divisor into working storage
  FOR I=N TO 0 STEP -1
    B3(I)=BCD._B(I)
  NEXT I
  
  'B3(N) contains first non-zero triplet, but which digit
  'within B3(N) is non zero
  J=0
  IF B3(N)>99 THEN
    J=3
  ELSE
    IF B3(N)>9 THEN J=2
    IF B3(N)<10 THEN J=1
  ENDIF
  'Number of significant digits in BCD.A()
  N=(N*3)+J
  
  'Do the Long Division
  FOR I=M-N TO 0 STEP -1
    
    'Calculate the Subtrahend for this iteration
    'Subtrahend=Divisor * 10 ^ I
    
    Cy=0
    J=10^I
    FOR K=0 TO A3L
      A4(K)=(B3(K)*J)+Cy
      Cy=INT(A4(K)/&H3E8)
      A4(K)=A4(K)-(&H3E8*Cy)  'a Mod n = a - (n * int(a/n)).
    NEXT K
    A4(K)=Cy
    
    FOR D=1 TO 9
      
      'Multiply the Divisor by D
      Cy = 0
      FOR K=0 TO A3L
        C1(K)=(A4(K)*D)+Cy
        Cy=C1(K)\&H3E8 
        C1(K)=C1(K) MOD &H3E8 
      NEXT K
      
      IF Cy=0 THEN
        J=K-1
      ELSE
        J=K
        C1(K)=Cy
      ENDIF
      
      'Subtract new divisor * Try digit from Dividend
      Bw=0
      FOR K =0 TO J
        C3(K)=A3(K)-C1(K)+&H3E8  - Bw
        Bw=1 -(C3(K)\&H3E8 )
        C3(K)=C3(K) MOD &H3E8 
      NEXT K
      
      'If No borrow then Dividend is greater than Subtractor
      IF Bw=0 THEN
        'Save away the result of the subtraction
        FOR K=0 TO J  'L
          Diff(K)=C1(K)
        NEXT K
      ELSE
        'If Borrow, then Dividend < Subtractor so use the last no-borrow subtraction & exit loop
        EXIT FOR
      ENDIF
      
    NEXT D
    
    'Update the Result
    C2(I)=D- 1
    'Update Length of Result
    C2L=C2L+ 1
    
    'If the multiplier was NON-Zero then we need to subtract the divisor*D from the dividend
    'ie calculate the remainder
    IF C2(I)>0 THEN
      
      Bw=0
      FOR K=0 TO A3L
        A5(K)=A3(K)-Diff(K)+&H3E8-Bw
        Bw= 1 -(A5(K)\&H3E8)
        A5(K)=A5(K) MOD &H3E8
      NEXT K
      
      'Find First non-zero triplet
      FOR K=Ele TO 0 STEP -1
        IF A5(K)<>0 THEN
          A3L=K
          EXIT FOR
        ENDIF
      NEXT K
      
      'Which becomes the new Dividend
      FOR K=0 TO A3L
        A3(K)=A5(K)
      NEXT K
      
    ENDIF
  NEXT I
  
  'Transfer the result from C2() to BCD.C()
  J=0
  FOR I=0 TO C2L-1 STEP 3
    BCD._C(J)=C2(I)+C2(I+1)*10+C2(I+2)*&H64          '&H64=100
    J=J+1
  NEXT I
  
  'Repack the Result from BCD._C() into BCD.C()
  J=0
  FOR I=0 TO (BCD.Size*2)-1 STEP 2
    BCD.C(J)=BCD._C(I) + BCD._C(I+1)*&H3E8 
    J=J+1
  NEXT I
  
END SUB
  
  
'=========================================================
'
'Clear BCD Register
'Dest is Element 0 of a BCD register, eg FREQ(0)
'NumDigits is Number of digits in the BCD register, is Optional
'
'The default Width is BCD.Size, but can be specified otherwise
'eg when clearing BCD.C() which is twice as long as BCD.A/B()
'
'This SUB is OPTIMISED for 24 digits
'
SUB BCD.Clr(Dest,Width)
  LOCAL I
  
  IF Width<1 THEN
    POKE VAR Dest,0,0:POKE VAR Dest,1,0:POKE VAR Dest,2,0:POKE VAR Dest,3,0
    POKE VAR Dest,4,0:POKE VAR Dest,5,0:POKE VAR Dest,6,0:POKE VAR Dest,7,0
    POKE VAR Dest,8,0:POKE VAR Dest,9,0:POKE VAR Dest,10,0:POKE VAR Dest,11,0
    POKE VAR Dest,12,0:POKE VAR Dest,13,0:POKE VAR Dest,14,0:POKE VAR Dest,15,0
  ELSE
    FOR I=0 TO (Width*4)-1
      POKE VAR Dest,I,0
    NEXT I
  ENDIF
  
END SUB
  

'=========================================================
'
'Copy a BCD string into a BCD Register/Destination
'Dest is the first element of a BCD register, eg BCD.A(0), CurFreq(0)
'A$ is a string of BCD digits, eg "1234567890"
'
SUB BCD.Load(Dest,A$)
  LOCAL I,J,M
  LOCAL B(BCD.Size),A2$(BCD.NumDigits)
  
  'Pad to make sure string is a multiple of 6 digits
  A2$(1)=BCD.LPad$(A$,BCD.Width)
  
  J = 0
  M=LEN(A2$(1))
  FOR I = M-BCD.Width+1 TO 1 STEP -BCD.Width
    B(J) = VAL(MID$(A2$(1), I, BCD.Width))
    J = J + 1
  NEXT I
  
  FOR I=0 TO (BCD.Size*4)-1
    POKE VAR Dest,I,PEEK(VAR B(0),I)
  NEXT I
  
  
END SUB
  
'=========================================================
'
'Prepend leading 0 to make sure
'string is a multiple of ML digits
'
'This SUB is OPTIMISED for 24 digits
FUNCTION BCD.LPad$(A$,ML)
  LOCAL I,J,M
  M=LEN(A$)
  
  IF M MOD ML <> 0 THEN
    FOR I=ML TO BCD.NumDigits STEP ML
      IF M<I THEN
        BCD.LPad$=STRING$(I-M,"0")+A$
        EXIT FOR
      ENDIF
    NEXT I
  ELSE
    BCD.LPad$=A$
  ENDIF
END FUNCTION
  
 
'==============================================
'
'Convert BCD.A() into an unsigned 32 bit integer
'in BCD.C()
'Algorithm process Sextets of decimal digits
'The Number must be less than 2^32, and can be at most 10 digits
'in Length, ie 2 Numeric vars with sextet packing
'The result will at most be 4 Numeric vars each storing 1 byte of the BIN
'
SUB BCD.A2Bin
  
  LOCAL I,L,Cy
  
  BCD.C(0)=0:BCD.C(1)=0:BCD.C(2)=0:BCD.C(3)=0
  
  'Each sextets of digits are worth 0-999999, ie 1,000,000's
  
  L=&HF4240         '1,000,000
  FOR I=1 TO 0 STEP -1
    
    'Add Next Decimal Digit into the Binary Long word
    BCD.C(0)=BCD.C(0)+(BCD.A(I)*L)
    
    'Has this caused an overflow
    IF BCD.C(0)>&HFF THEN
      'Yes
      'Now ripple the overflow across the other bytes
      Cy=BCD.C(0) AND &H7FFFFF00 
      BCD.C(0)=BCD.C(0)AND &HFF 
      BCD.C(1)=BCD.C(1)+(Cy\&H100)
      
      'Has this caused an overflow
      IF (BCD.C(1)>&HFF) THEN
        'Yes
        'Now ripple the overflow across the other bytes
        Cy=BCD.C(1) AND &H7FFFFF00 
        BCD.C(1)=BCD.C(1) AND &HFF 
        BCD.C(2)=BCD.C(2)+(Cy\&H100)
        
        'Has this caused an overflow
        IF (BCD.C(2)>&HFF) THEN
          'Yes
          'Now ripple the overflow across the other bytes
          Cy=BCD.C(2) AND &H7FFFFF00 
          BCD.C(2)=BCD.C(2) AND &HFF 
          
          BCD.C(3)=BCD.C(3)+(Cy\&H100)
          
        ENDIF
      ENDIF
    ENDIF
    
    L=1 
    
  NEXT I
  
  'Was there an overflow
  IF (BCD.C(3)>&HFF) THEN
    ERROR "OverFlow"
  ENDIF
  
END SUB
  
 
'=========================================================
'Returns a string representing the BCD Digits from BCD.A()
'eg if BCD.A(0)=123456, BCD.A(1)=789012,BCD.A(2)=345678, then
'BCD2ASC$ will return "345678789012123456"
'
'This SUB is OPTIMISED for 24 Digits
'
FUNCTION BCD.A2Asc$()
  
  LOCAL I,J,A$(1) Length 24,B$(1) Length 24
  
  FOR I=3 TO 0 STEP -1
    'Find First non Zero Sextet
    IF BCD.A(I)<>0 THEN
      FOR J=I TO 0 STEP -1
        'Uncomment the following line for Micromite code - faster
        B$(1)=B$(1)+STR$(BCD.A(J),BCD.Width,0,"0")
        'UnComment out the following 2 lines for Maximite
        '        A$(1)=STR$(BCD.A(J))
        '        B$(1)=B$(1)+STRING$(BCD.Width-LEN(A$(1)),"0")+A$(1)
      NEXT J
      'Now Remove any leading Zeroes
      FOR J=1 TO 4
        IF PEEK(VAR B$(1),J)<>&H30 THEN
          B$(1)=MID$(B$(1),J)
          EXIT FOR
        ENDIF
      NEXT J
      EXIT FOR
    ENDIF
  NEXT I
  
  BCD.A2Asc$=B$(1)
  
END FUNCTION
  
  
'=============================================================
'
'If C=1 then Double BCD.Size output
'
'
SUB BCD.Print(Src,Mode,C)
  
  LOCAL I,J,K
  
  IF C>0 THEN
    J=(BCD.Size*2)
  ELSE
    J=(BCD.Size)
  ENDIF
  
  LOCAL OutBCD(J)
  
  FOR I=0 TO (J*4)-1
    
    POKE VAR OutBCD(0),I,PEEK(VAR Src,I)
    
  NEXT I
  
  '  PRINT "BCD.Print:"; 'DEBUG
  FOR I=J-1 TO 0 STEP -1
    IF Mode=0 THEN
      PRINT STR$(OutBCD(I),BCD.Width,0,"0");" ";
    ELSE
      PRINT HEX$(OutBCD(I));" ";
    ENDIF
  NEXT I
  PRINT
END SUB


'=============================================================
'
' General purpose way to move BCD data between source and dest
' arrays. NOT as fast as doing a direct copy usin g a FOR loop.
'
' This SUB is OPTIMISED for 24 digits
'
SUB BCD.Copy(Src,Dest)

  POKE VAR Dest,0,PEEK(VAR Src,0):POKE VAR Dest,1,PEEK(VAR Src,1):POKE VAR Dest,2,PEEK(VAR Src,2):POKE VAR Dest,3,PEEK(VAR Src,3)
  POKE VAR Dest,4,PEEK(VAR Src,4):POKE VAR Dest,5,PEEK(VAR Src,5):POKE VAR Dest,6,PEEK(VAR Src,6):POKE VAR Dest,7,PEEK(VAR Src,7)
  POKE VAR Dest,8,PEEK(VAR Src,8):POKE VAR Dest,9,PEEK(VAR Src,9):POKE VAR Dest,10,PEEK(VAR Src,10):POKE VAR Dest,11,PEEK(VAR Src,11)
  POKE VAR Dest,12,PEEK(VAR Src,12):POKE VAR Dest,13,PEEK(VAR Src,13):POKE VAR Dest,14,PEEK(VAR Src,14):POKE VAR Dest,15,PEEK(VAR Src,15)
    
END SUB


'<<<<<<<<<<<<<<<<<<<<<<<<<Cut from here to end and Discard>>>>>>>>>>>>>>>>>>>>>>>

TestSuite:

  PRINT "Test Suite for BCD Arithmetic Routines"
  PRINT "The first set of digits per line is the execution time in mS"
  PRINT "of the BCD routine being tested."
  PRINT
  '
  
  IF MM.VER < 4.05 THEN
    ERROR "MM Basic V 4.5 or Newer is Required
    END
  ENDIF
  
  'Declare Global Vars here
  
  'Go RUN the main program
  Main
  
  PRINT
  PRINT "Test Suite Completed"
  
END
  
SUB Main()
  
  CPU 48
  
  LOCAL A$, B$, C$
  LOCAL I,J
  
  'Initialise the BCD Library
  'Configure for 24 decimal precision
  BCD.InitLib
  
  PRINT "Number of BCD Digits                    BCD.NumDigits:";BCD.NumDigits
  PRINT "Number of Decimal Digits per Numeric Var     BCD.Width";BCD.Width
  PRINT "Number of Numeric Vars per BCD Register       BCD.Size";BCD.Size
  PRINT
  
  'Load BCD.A using FOR Loop
  A$="789123456789654321"
  J=0
  L=TIMER
  FOR I=LEN(A$)-5 TO 1 STEP -6
    BCD.A(J)=VAL(MID$(A$,I,6))
    J=J+1
  NEXT I
  L=TIMER-L
  PRINT STR$(L,3,2);" Main:Load BCD.A using FOR Loop with ";A$;" BCD.A()=";
  BCD.Print BCD.A(0),0,0
  
  'Load BCD.A using Sub
  A$="123456789123456789654321"
  L=TIMER
  BCD.Load BCD.A(0),A$
  L=TIMER-L
  PRINT STR$(L,3,2);" Main:BCD.Load:BCD.A() with ";A$;" BCD.A()=";
  BCD.Print BCD.A(0),0
  
  'Clear BCD.A
  L=TIMER
  BCD.Clr BCD.A(0)
  L=TIMER-L
  PRINT STR$(L,3,2);" Main:BCD.Clr:BCD.A=";
  BCD.Print BCD.A(0)
  
  'Copy BCD.A to BCD.B
  A$="123456789"
  BCD.Load BCD.A(0),A$
  L=TIMER
  FOR I=0 TO BCD.Size-1
    BCD.B(I)=BCD.A(I)
  NEXT I
  L=TIMER-L
  PRINT STR$(L,3,2);" Main:Copy BCD.A to BCD.B using FOR loop:BCD.B()=";
  BCD.Print BCD.B(0)
  
  'Copy BCD.A to BCD.B using Sub
  BCD.Clr BCD.B(0)
  L=TIMER
  BCD.Copy BCD.A(0),BCD.B(0)
  L=TIMER-L
  PRINT STR$(L,3,2);" Main:Copy BCD.A to BCD.B using BCD.Copy:BCD.B()=";
  BCD.Print BCD.B(0)
  
  'Setup for BCD Add/Sub/Mul/Div tests
  A$="1030792140000000"
  L=TIMER
  BCD.Load BCD.A(0),A$
  L=TIMER-L
  PRINT STR$(L,3,2);" Main:BCD.Load:BCD.A() with ";A$;" BCD.A()=";
  BCD.Print BCD.A(0)
  
  B$="99999999"
  L=TIMER
  BCD.Load BCD.B(0),B$
  L=TIMER-L
  PRINT STR$(L,3,2);" Main:BCD.Load:BCD.B() with ";B$;" BCD.B()=";
  BCD.Print BCD.B(0)
  
  'Do BCD.C() = BCD.A() + BCD.B()
  B$="9999"
  BCD.Load BCD.B(0),B$
  L=TIMER
  FOR I=1 TO 100
    BCD.AddAB2C
  NEXT I
  L=TIMER-L
  PRINT STR$(L/100,3,2) ;" Main:BCD.AddAB2C of   ";A$;"+";B$;"=";
  BCD.Print BCD.C(0),0,0
  
  'Do BCD.A() = BCD.A() + BCD.B()
  A$="1030792140000000"
  BCD.Load BCD.A(0),A$
  B$="9999"
  BCD.Load BCD.B(0),B$
  L=TIMER
  FOR I=1 TO 100
    BCD.AddAB2A
  NEXT I
  L=TIMER-L
  PRINT STR$(L/100,3,2) ;" Main:BCD.AddAB2A of   ";A$;"+";B$;" x 100 =";
  BCD.Print BCD.A(0),0,0
  
  'Do BCD.C() = BCD.A() - BCD.B()
  B$="999"
  BCD.Load BCD.B(0),B$
  L=TIMER
  FOR I=1 TO 100
    BCD.SubAB2C
  NEXT I
  L=TIMER-L
  PRINT STR$(L/100,3,2) ;" Main:BCD.SubAB2C  :";A$;"-";B$;"=";
  BCD.Print BCD.C(0),0,0
  
  'Do BCD.A() = BCD.A() - BCD.B()
  A$="1030792140000000"
  BCD.Load BCD.A(0),A$
  B$="999"
  BCD.Load BCD.B(0),B$
  L=TIMER
  FOR I=1 TO 100
    BCD.SubAB2A
  NEXT I
  L=TIMER-L
  PRINT STR$(L/100,3,2) ;" Main:BCD.SubAB2A:  ";A$;"-";B$;" x 100 =";
  BCD.Print BCD.C(0),0,0
  
  'BCD.MulAB2C
  A$="34359738"
  B$="10000000"
  BCD.Load BCD.A(0),A$
  BCD.Load BCD.B(0),B$
  L=TIMER
  FOR I=1 TO 10
    BCD.MulAB2C
  NEXT I
  L=TIMER-L
  PRINT STR$(L/10,3,2);" Main:BCD.MulAB2C:";A$;"*";B$;"=";
  BCD.Print BCD.C(0),0,1
  
  'BCD.DivAB2C
  A$="4294967296"   '2^32
  B$="125"
  BCD.Load BCD.A(0),A$
  BCD.Load BCD.B(0),B$
  L=TIMER
  BCD.DivAB2C
  L=TIMER-L
  PRINT STR$(L,3,2);" Main:BCD.DivAB2C:";A$;"\";B$;"=";
  BCD.Print BCD.C(0),0,1
  
  'BCD.IncA
  A$="123456789"
  J=1
  BCD.Load BCD.A(0),(A$)
  L=TIMER
  FOR I=1 TO 100
    BCD.IncA J
  NEXT I
  L=TIMER-L
  PRINT STR$(L/100,3,2);" Main:BCD.IncA:";A$;"+" STR$(J*100);"=";
  BCD.Print BCD.A(0)
  
  'BCD.DecA
  A$="100000001"
  J=1
  BCD.Load BCD.A(0),(A$)
  L=TIMER
  FOR I=1 TO 100
    BCD.DecA J
  NEXT I
  L=TIMER-L
  PRINT STR$(L/100,3,2);" Main:BCD.DecA:";A$;"-" STR$(J*100);"=";
  BCD.Print BCD.A(0)
  
  'BCDA2BIN
  A$="343597380"
  BCD.Load BCD.A(0),(A$)
  L=TIMER
  FOR I=1 TO 10
    BCD.A2Bin
  NEXT I
  L=TIMER-L
  PRINT STR$(L/10,3,2);;" Main:BCD.A2Bin of ";A$;"=&H";
  BCD.Print BCD.C(0),1,0
  
  'BCDA2ASC
  A$="03043597380"
  BCD.Load BCD.A(0),(A$)
  L=TIMER
  FOR I=1 TO 10
    C$=BCD.A2Asc$()
  NEXT I
  L=TIMER-L
  PRINT STR$(L/10,3,2);" Main:BCD.A2Asc$ of ";A$;"=";C$
  
END SUB
  
